import sys
import errno
import socket
from struct import pack, unpack

from Crypto.Cipher import PKCS1_OAEP

from configuration import Configuration
import binascii

import time
import struct

from encryption import CipherHelper

'''
This package sends a *complete* data package, or fails.
To do this, the size of the data to send is added to the package, so that
when recieving, we know if we got all the data or just a bit.
'''

'''
SecureCommunication

First package from N1 to N2:
send to N2: N2_RSAPUBKEY('GREETINGS!'+AES,IV)   # AES&IV from N1

N2 decrypts msg, if comprehensible (e.g it starts with 'GREETINGS!'), extract AES keypair
N2 responds with a, at least 32 bytes, meassage: 'FOR THE PEOPLE'+DATETIME or some random junk
N1 now knows N2 is authentic (or it is u to N2 to be) and new we can communicate securely

'''


# todo: make the RSA a one-shot read, and the AES streamed, so we can encrypt the size variable too


# Class for initiating secure communication with another node
# Connects to a node with a GREETINGS! query, encrypted with the nodes bublic rsa key
# Also an AES keypair is sent over which now will be used for communication
class SecureCommunication:
    def __init__(self, ip, port, other_rsa_pub_key):
        self.ch = CipherHelper()

        # Connect to other node
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.socket.connect((ip, int(port)))
        except:
            self.socket = None
            return
        query = 'GREETINGS!' + self.ch.aes_key.hex() + ',' + self.ch.raw_iv.hex()

        # crypt with rsa pubkey from other node:
        cipher = PKCS1_OAEP.new(other_rsa_pub_key)
        rsa_query = cipher.encrypt(str.encode(query))  # encode to bytes

        # Send RSA encrypted message
        if socket_secure_send(self.socket, rsa_query):
            answer = socket_secure_recv(self.socket, 0)
            if answer is not None:
                # If everything went well, the other node have encrypted the massage with AES
                message = self.ch.decrypt(answer)
                if message != b'HI_MINDOKI':
                    print('SC connection refused')
                    self.socket = None
                else:
                    print('SC Securely connected to other node.')
            else:
                print('SC Unknown answer / error')

                self.socket = None
        else:
            self.socket = None

    # True if link is open and working
    def is_active(self):
        return self.socket is not None

    # Sends a message and returns the recv
    def send(self, msg):
        if not self.is_active():
            return None

        if socket_secure_send(self.socket, self.ch.encrypt(msg)):
            data = socket_secure_recv(self.socket, 0)
            if data is not None:
                return self.ch.decrypt(data)
            else:
                print('Could not recv after request was sent')
        else:
            print('Could not send request')
        return None

    def close(self):
        if self.socket is not None:
            self.socket.close()
            self.socket = None


# This is the pair class with SecureCommunication but at the listening part
# Instantiate one when someone tries to connect to the listener
# It will get the AES key pair etc.
class SecureCommunicationListen:
    def __init__(self, socket, my_rsa_priv_key, max_read_size):
        self.aes_key = None
        self.aes_iv = None
        self.socket = socket

        incoming = socket_secure_recv(self.socket, 800)  # Recv at least 1 RSA block (AES (key+iv) + 10 char handshake + padding)
        # Reading finalized

        # Avoid catching normal network errors:
        if incoming is not None:
            # decrypt with our private RSA key
            try:
                cipher = PKCS1_OAEP.new(my_rsa_priv_key)
                message = cipher.decrypt(incoming)
            except ValueError:
                print("#Error: SCL could not connect to peer, decrypting raised ValueError:", ValueError)
                self.socket.close()
                self.socket = None
                return

            # Split up the query in requirement and payload data
            query = message[:10]  # Query is always 10 bytes
            data = message[10:]  # the payload

        # check for the initial greeting/handshake command:
        if query != b'GREETINGS!':
            print('SCL Did not get greeting message, closing down link')
            self.socket.close()
            self.socket = None
            return

        # Payload should contain AES key and initialisation vector:
        tmp = data.split(b',')

        if len(tmp) != 2:  # error
            self.socket.close()
            self.socket = None
            return

        key = tmp[0]
        iv = tmp[1]

        # split out AES, key is in bytes, IV is an int
        self.ch = CipherHelper(aes_key=binascii.unhexlify(key), raw_iv=binascii.unhexlify(iv))

        # return OK message, encrypted with AES
        rmsg = self.ch.encrypt('HI_MINDOKI'.encode())

        # send back handshake ACK AES(HI_MINDOKI)
        socket_secure_send(self.socket, rmsg)

    # True if link is open and working
    def is_active(self):
        return self.socket is not None

    # Get queries and data from caller
    def recv(self):
        if not self.is_active():
            return None
        data = socket_secure_recv(self.socket, 0)
        if data is not None:
            return self.ch.decrypt(data)
        else:
            print('SCL Could not recv after request was sent')
            return None

    # Send msg and data back to caller
    def send(self, msg):
        if not self.is_active():
            return None

        if socket_secure_send(self.socket, self.ch.encrypt(msg)):
            data = socket_secure_recv(self.socket, 0)
            if data is not None:
                return self.ch.decrypt(data)
            else:
                print('Listener: Could not recv after request was sent')
        else:
            print('Listener: Could not send request')
        return None

    def close(self):
        if self.socket is not None:
            self.socket.close()
            self.socket = None


# Helpers for sending a specific sized package:


# Encodes a long long (8bytes) for the size of the rest of the data
def socket_secure_send(s, msg, extensive_logging=False):
    if extensive_logging:
        print('SSS send message')
        print('msg type', type(msg), 'should be binary)')

    # As we all know, data sent over the internet might be split up, so send how many
    # bytes we'll try to push through here. An unsigned 8Byte integer should do the trick
    # hopefully for the foreseeable future:
    datasize = len(msg)
    lengthdata = pack('>Q', datasize)
    # Send a long long, little endian encoded information about the msg size
    try:
        # Send size of data to come
        if extensive_logging:
            print('SSS Send length data', len(lengthdata),' should be 8, extracted datasize to receive=', datasize)
        s.send(lengthdata)

        # Send off the actual data
        if extensive_logging:
            print("SSS Send data of size", lengthdata)

        send_by_packages = True

        if send_by_packages:  # Use this!
            if extensive_logging:
                print('SSS --- Send data by packages Start ---')

            # Get config information
            config = Configuration()

            # send by packets
            max_packet_size = config.get('SERVER_SEND_PACKAGE_SIZE', 1024 * 32)
            data_size = len(msg)
            left_to_send = data_size
            sent = 0
            while left_to_send > 0:
                package_size = min(max_packet_size, left_to_send)
                if extensive_logging:
                    print('SSS bytes left to send', left_to_send, ', send a data package of max size', package_size)
                    #print(struct.unpack("B" * 7 + "I" * 24, s.getsockopt(socket.SOL_TCP, socket.TCP_INFO, 104)))

                # What? Didn't we already send the size of the data earlier?
                # We sure did, but this is so when someone is downloading your 1.5TB of
                # knitting recipes and videos, it won't choke the OS:s small buffers!
                # We wouldn't want that right? So here we'll use a smaller 32bits unsigned integer,
                # but beware, buffers can be small so don't use numbers too big!
                lengthdata = pack('>L', package_size)
                if extensive_logging:
                    print('send length data:')
                s.send(lengthdata)
                if extensive_logging:
                    print('length data sent.')

                # Now send a chunk of the data
                data_to_send = msg[sent:sent+package_size]
                data_to_send_len = len(data_to_send)
                if extensive_logging:
                    print('send(): len=' + str(data_to_send_len)+')')
                a = s.send(data_to_send)
                if extensive_logging:
                    print('send(' + str(len(data_to_send))+') Done, returned:', a)

                sent += package_size
                left_to_send -= package_size

                # And wait for the little Ack!
                if extensive_logging:
                    print('wait for Chunk-ACK...')
                ack = s.recv(1)  # todo check the ack = b'1' or something
                if extensive_logging:
                    print('Got Chunk-ACK:', ack)
            if extensive_logging:
                print('SSS --- Send data by packages Done ---')

        else:  # This is a Bad idea (tm) except for dinky data
            if extensive_logging:
                print('SSS sending', len(msg), 'bytes of data, in one go.')
            s.send(msg)

        # Check if the data went through, other socket sends b'0' for success and others for error
        if extensive_logging:
            print("SSS recv ACK")

        if extensive_logging:
            print('wait for (global) ACK...')
        ack = s.recv(1)
        if extensive_logging:
            print("SSS Got ACK:", ack)

        return ack == b'0'
    except socket.error:
        print("SSS There was a problem in socket_secure_send")

    return False


def socket_secure_recv(s, max_read_size, extensive_logging=False):
    # Recv the long long size data
    if extensive_logging:
        print("SSR Recv length data")

    try:
        tmp = s.recv(8)
    except socket.error as e:
        err = e.args[0]
        if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
            if extensive_logging:
                print('SSR no (size) data avaliable')

            # No data available
            return None
        else:
            # Real error:
            print("SSR error: ", e)

            return None

    if len(tmp) != 8:
        if extensive_logging:
            print('SSR Warning: Socket probably closed on other side: Could not read size-variable from incoming stream, len=',len(tmp))
        return None

    (to_read, ) = unpack('>Q', tmp)
    
    if extensive_logging:
        print("SSR Recv length data, expected size to read:", to_read)

    # Log progress if file is bigger then 16KB
    log_progress = to_read >= 16384

    # We are now supposed to read to_read bytes of data, bail out if too big
    if max_read_size != 0 and to_read > max_read_size:
        if extensive_logging:
            print("SSR Wrong size / too much data, max_read_size=", max_read_size, " to_read=", to_read, " max_read_size=", max_read_size)

        s.send(b'1')
        return None
    data = b''
    read = to_read

    recv_by_packages = False

    if recv_by_packages:  # Don't use this

        # data is sent by packets (so we won't try to get a TB in one go)
        max_packet_size = 3276800
        data_size = to_read
        left_to_recv = data_size
        done = 0
        while left_to_recv > 0:
            package_size = min(max_packet_size, left_to_recv)
            if extensive_logging:
                print('SSR bytes left to recv', left_to_recv, ' so recv a data package of size', package_size, ' total recvd:', done)
                print(struct.unpack("B"*7+"I"*24, s.getsockopt(socket.SOL_TCP, socket.TCP_INFO, 104)))
            part = s.recv(package_size)
            if part is None:
                print('Error in receiving data...')
                s.close()
                return
            data += part
            received = len(part)
            left_to_recv -= received
            done += received
            time.sleep(0.2)
    else:  # Use this!
        if extensive_logging:
            print('SSR --- Receive data Start ---')

        while read > 0:
            # Read data
            if extensive_logging:
                if read == to_read:
                    print('SSR read some data, wanted size=', read)  # When we haven't read any data
                else:
                    print('SSR missing data', read, 'so recv some more, already read:', len(data))
            # print(struct.unpack("B"*7+"I"*24, s.getsockopt(socket.SOL_TCP, socket.TCP_INFO, 104)))

            # Recv our first chunk:
            chunk_size_data = s.recv(4)  # A 4 bit unsigned int for chunk size
            if len(chunk_size_data) != 4:
                if extensive_logging:
                    print('SSR Warning: Socket probably closed on other side: Could not get shunk size data')
                return None

            (chunk_size,) = unpack('>L', chunk_size_data)
            if extensive_logging:
                print('-- Read chunk  (chunk size=' + str(chunk_size) + ') --')
            while chunk_size > 0:
                if extensive_logging:
                    print(' left to read:', (to_read - len(data)), 'Try to recv...')
                part = s.recv(min(to_read, 100000000))  # don't try to read too much in one go (WIP, try smaller numbers when it works already)
                if extensive_logging:
                    print('recv:d ' + str(len(part)) + ' bytes')
                if part is None:
                    if extensive_logging:
                        print('Error in receiving data...')
                    s.close()
                    return

                data += part
                chunk_size -= len(part)

                # Show the download progress (if over 16kb)
                if log_progress:
                    percent = 100 - int(((to_read - len(data)) * 100) / to_read)
                    print('Downloading File [%d%%]\r' % percent, end="", flush=True)

                # send ACK
            if extensive_logging:
                print('-- All chunks read Done --')
                print('Send chunk-Ack:')
            s.send(b'0')

            read = to_read - len(data)

            # Show the download progress (if over 16kb)
            if log_progress:
                percent = 100 - int((read*100) / to_read)
                print('Downloading File [%d%%]\r'%percent, end="", flush=True)

        if log_progress:
            if read <= 0:
                print('Downloading File done [100%]');
            else:
                print('Error downloading file...')

        if extensive_logging:
            print('SSR --- Receive data Done ---')


    # Reading finalized
    if extensive_logging:
        print("SSR Send ACK")

    s.send(b'0')
    return data
